import * as d3 from 'd3';
import { NetlistRender } from '.';
import { globalSetting } from '../global';

export class InstantiationRender {
    /**
     * 
     * @param {d3.Selection} selection 
     * @param {NetlistRender} rootRender 
     */
    constructor(selection, rootRender) {
        this.parentSelection = selection;
        this.rootRender = rootRender;

        /**
         * @type {BasicD3DataItem[]}
         */
        this.data = []

        /**
         * @description id 到管理数据项的映射
         * @type {Map<string, BasicD3ManagmentItem[]>}
         */
        this.id2manager = rootRender.id2manager;
    }

    /**
     * @description 将 elknode 关于 例化模块 的数据添加为 d3 数据项目
     * @param {ElkNode} node 
     */
    addAsD3DataItem(node) {        
        this.data.push({
            id: node.id,
            x: node.x,
            y: node.y,
            name: node.name,
            width: node.width,
            height: node.height,
            fill: 'var(--main-dark-color)',
            text: node.renderName,
            rx: 3,
            ry: 3
        });
    }

    render() {
        const data = this.data;
        const rootRender = this.rootRender;
        const id2manager = this.id2manager;
        const _this = this;

        let instantiationSelections = this.parentSelection.selectAll('g.instance')
            .data(data)
            .enter()
            .append('g')
            .attr('class', 'instance')
            .attr("transform", d => `translate(${d.x}, ${d.y})`);
        
        let instances = instantiationSelections.append('rect')
            .attr('width', data => data.width)
            .attr('height', data => data.height)
            .attr('fill', d => d.fill);
        
        let texts = instantiationSelections.append('text')
            .attr('x', data => data.width / 2)              // 文本的 x 坐标（居中）
            .attr('y', data => data.height / 2)             // 文本的 y 坐标（居中）
            .attr('dominant-baseline', 'middle')            // 文本垂直居中
            .attr('text-anchor', 'middle')                  // 文本水平居中
            .attr('fill', 'var(--foreground)')              // 文本颜色
            .attr('font-size', '0')
            .transition()
            .duration(1000)
            .attr('font-size', '15px')
            .attr('class', 'port-caption')
            .text(data => data.name);                       // 设置文本内容

        
        if (globalSetting.renderAnimation) {
            instances.transition()
                .duration(1000)
                .attr('stroke', 'var(--main-color)')
                .attr('stroke-width', 2)
                .attr('rx', d => d.rx)
                .attr('ry', d => d.ry);
        } else {
            instances.attr('stroke', 'var(--main-color)')
                .attr('stroke-width', 2)
                .attr('rx', d => d.rx)
                .attr('ry', d => d.ry);
        }

        instantiationSelections
            .attr('class', 'grab')
            .each(function (data) {                    
                const portSelection = d3.select(this);
                const manager = _this.createDataManager(portSelection, data);

                // TODO: 实现拖拽
                // registerDragEvent(manager, rootRender);       
            });
            
        this.selections = instantiationSelections;
        return instantiationSelections;
    }

    /**
     * 
     * @param {d3.Selection} selection 
     * @param {BasicD3DataItem} data 
     * @returns {BasicD3ManagmentItem}
     */
    createDataManager(selection, data) {
        const id2manager = this.id2manager;
        // 创建拖拽上下文
        const dragContext = {
            neighbors: [],
            elkGraph: {
                // elk 是无状态的，id 取什么名字都行
                id: 'root',
                children: [],
                edges: [],
                layoutOptions: {}
            }
        }
        const managerItem = {
            data,
            selection,
            type: 'cell',
            dragContext
        };

        if (!id2manager.has(data.id)) {
            id2manager.set(data.id, []);
        }

        id2manager.get(data.id).push(managerItem);
        return managerItem;
    }
}