/**
 * 将各个节点转换为 elknode 的数据结构
 */

import { globalLookup } from "../global";
import { Cell, dotConnect, ModuleView } from "./yosys";

export const LINE_WIDTH = 2;

export const LAYOUT_CONSTANT = {
    // port
    PORT_WIDTH: 50,
    PORT_HEIGHT: 20,
    
    // 例化模块（这只是最小，height 会根据端口数量进行调整）
    INSTANTIATION_WIDTH: 80,
    INSTANTIATION_HEIGHT: 80,

    // port 顶部的空余，一般用来放置标签的
    PORT_TOP_MARGIN: 30,

    // 常数
    CONSTANT_WIDTH: 50,
    CONSTANT_HEIGHT: 50,

    // 器件的端口
    CELL_PORT_HEIGHT: 1,
    CELL_PORT_WIDTH: 1
};

export const ELK_DIRECTION = {
    LEFT: 'WEST',
    RIGHT: 'EAST',
    TOP: 'NORTH',
    BOTTOM: 'SOUTH'
}


export function makeEdgeId(fromId, toId) {
    return fromId + '-' + toId;
}

export class Module {
    /**
     * @param {string} name
     * @param {YosysNetModule} module 
     */
    constructor(name, module) {
        this.module = module;
        this.name = name;

        /**
         * @description
         * @type {ModuleView}
         */
        this.view = new ModuleView(name, module);
    }
    
    /**
     * @description 从 ports 中创建所有变量相关的节点
     * @returns {ElkNode[]}
     */
    makeNetsElkNodes() {
        const nodes = [];

        // 绘制 ports
        for (const name of this.view.nameToPort.keys()) {
            const port = this.view.nameToPort.get(name);


            if (port.direction === 'input') {
                // 为 port 设置连接点
                const portConnection = {
                    id: dotConnect(port.id, '0'),
                    renderType: 'portConnection',
                    width: 1,
                    height: 1,
                    x: LAYOUT_CONSTANT.PORT_WIDTH,
                    y: LAYOUT_CONSTANT.PORT_HEIGHT / 2
                };
                
                const node = {
                    id: port.id,
                    name: port.name,
                    renderName: name,
                    renderType: 'port',
                    type: port.direction,
                    width: LAYOUT_CONSTANT.PORT_WIDTH,
                    height: LAYOUT_CONSTANT.PORT_HEIGHT,
                    ports: [portConnection],
                    layoutOptions: {
                        'partitioning.partition': 1,
                        'org.eclipse.elk.portConstraints': 'FIXED_POS'
                    }
                };
    
                nodes.push(node);
            } else {
                // 为 port 设置连接点
                const portConnection = {
                    id: dotConnect(port.id, '0'),
                    renderType: 'portConnection',
                    width: 1,
                    height: 1,
                    x: 0,
                    y: LAYOUT_CONSTANT.PORT_HEIGHT / 2
                };

                const node = {
                    id: port.id,
                    name: port.name,
                    renderName: name,
                    renderType: 'port',
                    type: port.direction,
                    width: LAYOUT_CONSTANT.PORT_WIDTH,
                    height: LAYOUT_CONSTANT.PORT_HEIGHT,
                    ports: [portConnection],
                    layoutOptions: {
                        'partitioning.partition': 999,
                        'org.eclipse.elk.portConstraints': 'FIXED_POS'
                    }
                };
    
                nodes.push(node);
            }

        }

        // 非 port 的其他内部变量的节点不需要绘制，因为它们总能被表示为中间变量
        // 它们的操作可以表示为一系列器件的组合或者直连

        return nodes;
    }

    /**
     * @description 从 cells（器件） 中创建节点
     * @returns {[ElkNode[], ElkEdge[]]}
     */
    makeCellsElkNodes() {
        const nodes = [];
        const edges = [];

        const skinManager = globalLookup.skinManager;
        for (const name of this.view.nameToCell.keys()) {
            const cell = this.view.nameToCell.get(name);
            const skin = skinManager.querySkin(cell.type);

            if (skin) {
                // 内置器件
                // 创建器件节点的 port, port 和 connection 一一对应
                const meta = skin.meta;
                
                const height = meta.height;
                const width = meta.width;

                const ports = [];

                // 统计分配到左右两侧的 port
                const leftSideConnections = [];
                const rightSideConnections = [];

                for (const conn of cell.nameToConnection.values()) {
                    if (conn.direction === 'input') {
                        leftSideConnections.push(conn);
                    } else {
                        rightSideConnections.push(conn);
                    }
                }

                // 计算左侧的
                for (let i = 0; i < leftSideConnections.length; ++ i) {
                    const connection = leftSideConnections[i];
                    const yOffset = meta.getPortYOffset(connection.name);

                    ports.push({
                        id: connection.id,
                        renderName: connection.name,
                        renderType: 'cellPort',
                        width: LAYOUT_CONSTANT.CELL_PORT_WIDTH,
                        height: LAYOUT_CONSTANT.CELL_PORT_HEIGHT,
                        x: 0,
                        y: yOffset
                    });


                }

                // 计算右侧的
                for (let i = 0; i < rightSideConnections.length; ++ i) {
                    const connection = rightSideConnections[i];
                    const yOffset = meta.getPortYOffset(connection.name);                  

                    ports.push({
                        id: connection.id,
                        renderName: connection.name,
                        renderType: 'cellPort',
                        width: LAYOUT_CONSTANT.CELL_PORT_WIDTH,
                        height: LAYOUT_CONSTANT.CELL_PORT_HEIGHT,
                        x: width,
                        y: yOffset
                    });
                }

                const node = {
                    id: cell.id,
                    name: cell.name,
                    renderName: cell.type,
                    renderType: 'cell',
                    width,
                    height,
                    ports,
                    layoutOptions: {
                        'org.eclipse.elk.portConstraints': 'FIXED_POS',
                        'partitioning.partition': 10
                    }
                };
                nodes.push(node);
            } else {
                // 例化模块
                    
                // 创建器件节点的 port, port 和 connection 一一对应
                const ports = [];                
                for (const connectionName of cell.nameToConnection.keys()) {                    
                    const connection = cell.nameToConnection.get(connectionName);                
                    const portSide = connection.direction === 'input' ? ELK_DIRECTION.LEFT: ELK_DIRECTION.RIGHT;
                    ports.push({
                        id: connection.id,
                        renderName: connectionName,
                        renderType: 'cellPort',
                        width: LAYOUT_CONSTANT.CELL_PORT_WIDTH,
                        height: LAYOUT_CONSTANT.CELL_PORT_HEIGHT,
                        properties: {
                            'port.side': portSide
                        }   
                    })
                }

                const node = {
                    id: cell.id,
                    name: cell.name,
                    renderName: cell.type,
                    renderType: 'cell',
                    width: LAYOUT_CONSTANT.INSTANTIATION_WIDTH,
                    height: LAYOUT_CONSTANT.INSTANTIATION_HEIGHT,
                    ports,
                    layoutOptions: {
                        // 强制固定 port 的方向
                        'org.eclipse.elk.portConstraints': 'FIXED_SIDE',
                        'partitioning.partition': 10
                    }
                };

                nodes.push(node);
            }
        }

        return nodes;
    }

    /**
     * @description 从 cells.connections 中创建边
     * @returns {[ElkNode[], ElkEdge[]]}
     */
    makeConnectionElkNodes() {
        const nodes = [];
        const edges = [];

        const tree = this.view

        for (const cellName of tree.nameToCell.keys()) {
            const cell = tree.nameToCell.get(cellName);
            
            for (const connectionName of cell.nameToConnection.keys()) {
                const connection = cell.nameToConnection.get(connectionName);
                // 遍历器件端口的每一个连接点
                // 比如对于端口 input [31:0] data ，它的32个位不一定是完全导向一个变量（虽然我们愿意认为，大部分情况下是这样）
                // 大部分情况下，👇的 wireIds.length 都是 1 
                for (const wire of connection.wires) {
                    const wireId = wire.wireId;
                    const id = wire.id;
                    if (typeof wireId === 'string') {
                        // 常数
                        // 如果是常数，需要先创建代表常数的节点，常数一定是器件的输入，而非输出
                        const node = {
                            id,
                            name: wireId,
                            width: LAYOUT_CONSTANT.CONSTANT_WIDTH,
                            height: LAYOUT_CONSTANT.CONSTANT_HEIGHT,
                            layoutOptions: {
                                'elk.port.side': 'WEST',
                                'partitioning.partition': 10
                            }
                        };

                        nodes.push(node);

                        // 创建常数到器件的连线
                        const edge = {
                            id: makeEdgeId(cell.id, id),
                            source: cell.id,
                            sourcePort: connection.id,
                            target: id,
                            targetPort: id
                        };

                        edges.push(edge);
                    } else {
                        // 如果不是一个常数连接，那么存在两种可能
                        // 1. 当前的器件的这个端口和某一个 port 连接
                        // 2. 当前的器件的这个端口和另一个器件的一个端口连接
                        if (tree.wireIdToPort.has(wireId)) {

                            // 当前的器件的这个端口和某一个 port 连接
                            const port = tree.wireIdToPort.get(wireId);

                            if (port.direction === 'input') {
                                const edge = {
                                    // id 遵循 sourcePort-targetPort
                                    id: makeEdgeId(port.id, connection.id),
                                    source: port.id,
                                    sourcePort: dotConnect(port.id, '0'),
                                    target: cell.id,
                                    targetPort: connection.id
                                };
                                
                                edges.push(edge);
                            } else {
                                const edge = {
                                    id: makeEdgeId(connection.id, port.id),
                                    source: cell.id,
                                    sourcePort: connection.id,
                                    target: port.id,
                                    targetPort: dotConnect(port.id, '0')
                                };    
                                edges.push(edge);
                            }
                        } else if (tree.wireIdToConnection.has(wireId)) {
                            
                            // 当前的器件的这个端口和另一个器件的一个端口连接
                            const conn = tree.wireIdToConnection.get(wireId);

                            // 防止查询到自己
                            if (conn.id === connection.id && conn.cell.id === cell.id) {
                                continue;
                            }

                            if (conn.direction === 'input') {
                                const edge = {
                                    // id 遵循 sourcePort-targetPort
                                    id: makeEdgeId(conn.id, connection.id),
                                    source: conn.cell.id,
                                    sourcePort: conn.id,
                                    target: cell.id,
                                    targetPort: connection.id
                                };
    
                                edges.push(edge);
                            } else {
                                const edge = {
                                    id: makeEdgeId(connection.id, conn.id),
                                    source: cell.id,
                                    sourcePort: connection.id,
                                    target: conn.cell.id,
                                    targetPort: conn.id
                                };
                                edges.push(edge);
                            }
                        }
                    }
                }

            }
        }

        return [nodes, edges];
    }


}
