import * as d3 from 'd3';
import { globalSetting } from '../global';
import { NetlistRender } from '.';

export class WireRender {
    /**
     * 
     * @param {d3.Selection} selection
     * @param {NetlistRender} rootRender  
     */
    constructor(selection, rootRender) {
        this.selection = selection;
        this.rootRender = rootRender;
        
        /**
         * @type {BasicD3DataItem[]}
         */
        this.data = [];

        /**
         * @description id 到管理数据项的映射
         * @type {Map<string, BasicD3ManagmentItem[]>}
         */
        this.id2manager = rootRender.id2manager;
    }

    /**
     * @description 将 elknode 关于 wire 的数据添加为 d3 数据项目
     * @param {ElkPoint[]} points 
     * @param {ElkEdge} edge 
     */
    addAsD3DataItems(points, edge) {
        for (let i = 0; i < points.length - 1; ++ i) {
            this.data.push({
                id: edge.id,
                x1: points[i].x,
                y1: points[i].y,
                x2: points[i + 1].x,
                y2: points[i + 1].y,
                strokeWidth: 2,
                color: 'var(--foreground)'
            });
        }
    }

    render() {
        const data = this.data;
        const id2manager = this.id2manager;
        const _this = this;

        let lineSelections = this.selection.selectAll('line')
            .data(data)
            .enter()
            .append('line')
            .attr('x1', data => data.x1)
            .attr('y1', data => data.y1)
            .attr('x2', data => data.x2)
            .attr('y2', data => data.y2)
            .attr('stroke', data => data.color);
        
        if (globalSetting.renderAnimation) {
            lineSelections = lineSelections
                .transition()
                .duration(1000);
        }

        lineSelections
            .attr('stroke-width', data => data.strokeWidth)
            .each(function (data) {
                const selection = d3.select(this);
                const manager = _this.createDataManager(selection, data);
            });
    }

    /**
     * 
     * @param {d3.Selection} selection 
     * @param {BasicD3DataItem} data 
     * @returns {BasicD3ManagmentItem}
     */
    createDataManager(selection, data) {
        const id2manager = this.id2manager;
        // wire 不需要拖拽上下文
        const managerItem = {
            data,
            selection,
            type: 'wire'
        };

        if (!id2manager.has(data.id)) {
            id2manager.set(data.id, []);
        }

        id2manager.get(data.id).push(managerItem);
        return managerItem;
    }
}